// Estimate TiDeH parameters from Twitter data:
//    V.160101  by Ryota Kobayashi 
#include "TiDeH.h" 

//  Cnt. #Retweet in a window
int     cnt_rt( double t_st, double t_en, struct data_rt data );
//  Estimate p(t) from data by moving window.
double  est_pt( double t_st, double t_en, struct data_rt data, double s0 );

// Fit damping oscillating function.
double  fit_exp_LS( int num, const double x[], const double p_0[], double par[], double q[]); 
// Initialize parameters
void  init_par( int num, const double p_0[], double par[] );


// Lovenberg-Marquadt method
double  solv_LM( int num, const double x[], const double p_0[], double *c_LM, double par[], const double q[] ); 
// Calc. Matrix & Vector in the Gauss-Newton method.
double calc_FH( int num, const double x[], const double p_0[], const double th[], const double q[], double z[], double H[] );


int main(int argc, char *argv[])
{ 
  if( argc != 5 )   
    {   printf("Usage:./a.out *f_in *f_out $s0  $T0 (h)\n");  exit(1);   }
  
  FILE *f_in, *f_out; 
  f_in  = fopen( argv[1], "r");  
  if ( f_in == NULL)    {  printf("Cannot open $f_in\n");   exit(1);  }
  f_out = fopen( argv[2], "w");  
  if ( f_out == NULL)   {  printf("Cannot open $f_out\n");  exit(1);  }
  
  int    i, num;     double  t_0;
  struct data_rt   data;
  
  fscanf( f_in, "%d %lf", &num, &t_0);    // Read #Event
  data.n_i = calloc( num, sizeof(int) );     data.num= num;
  data.t_sp= calloc( num, sizeof(double) );  
  
  for (i=0; i<num; i++)   // Read event time and #followers. 
    {  fscanf( f_in, "%lf  %d ", &data.t_sp[i], &data.n_i[i]);  }
  
  int      dt= 1, max= atoi(argv[4]), *n_d;
  double   *p_t, *x, s0= atof(argv[3]), t_st, t_en;    
  n_d= calloc( max-WIN+1, sizeof(int) ); 
  p_t= calloc( max-WIN+1, sizeof(double) );     
  x= calloc( max-WIN+1, sizeof(double) );
  
  for (i=0; i< max-WIN+1; i++)
    {
      t_st= dt*i;     t_en= dt*i+ WIN;      
      
      x[i]= (t_st+t_en)/ 48.0;
      n_d[i]= cnt_rt( t_st, t_en, data); 
      p_t[i]= est_pt( t_st, t_en, data, s0);             
    }

  double   err, par[DIM], q[2];   // q[0]: t_0, q[1]: omega.  
  q[0]= t_0;   q[1]= 2* M_PI;
  err= fit_exp_LS( max-WIN+1, x, p_t, par, q);
  fprintf ( f_out, "%f %f %f %f \n", par[0], par[1], par[2], par[3]); 
  
  free(data.t_sp);     free(data.n_i);    free(p_t); 
  fclose(f_in);        fclose(f_out);     return 0;        
}


int     cnt_rt( double t_st, double t_en, struct data_rt data )
{
  int      k= 0, n_rt= 0;   
    
  while ( data.t_sp[k]< t_en && k< data.num ) // k-th event    
    {   // Cnt. #Events in a window      
      if ( k> 0 &&  data.t_sp[k] > t_st )   {  n_rt++;  }	      
      k++;
    } 

  return n_rt;
}


//  Estimate p(t) from data by moving window.
double  est_pt( double t_st, double t_en, struct data_rt data, double s0 )
{  
  int      k= 0, n_rt= 0;   
  double   c0= 1.0/s0/ (1- 1.0/Th), h1, h2, par, sum= 0;
    
  while ( data.t_sp[k]< t_en && k< data.num ) // k-th event    
    {   
      // h_t= \int_0^t \lam_s ds
      h2= calc_ht( (double)(t_en- data.t_sp[k]), c0, s0 );
      h1= calc_ht( (double)(t_st- data.t_sp[k]), c0, s0 );
      // sum: ¥sum_k ¥int ¥lambda(s) ds   
      sum += data.n_i[k]* (h2- h1); 

      // Cnt. #Events in a window
      if ( k> 0 &&  data.t_sp[k] > t_st )   {  n_rt++;  }	      
      k++;
    } 

  // par: p_T
  if ( n_rt== 0 )   {  par= 0;   }    
  else              {  par=  n_rt/sum;    }
  
  return par;
}


// Fit damping oscillating function.
double  fit_exp_LS( int num, const double x[], const double p_0[], double par[], double q[])
{
  int     n_r= 0;
  double  c_LM= 0.001, eps= 1.0;     
  init_par( num, p_0, par);  
  
  while ( eps> pow(0.1, 6)   )   // Maximize Likelihood 
    { 
      eps= solv_LM( num, x, p_0, &c_LM, par, q);     n_r++;  
      if( n_r> pow(10, 4)  )
        {  printf("# Error %f, %f, %f\n", eps, par[0], par[1]);   break;  }
    }   
  // Error
  double  F_x[DIM], J_x[DIM*DIM];
  return    calc_FH( num, x, p_0, par, q, F_x, J_x);
}

// Initialize parameters
void  init_par( int num, const double p_0[], double par[] )
{
  int   i;     double   ave= 0;
  
  for (i=0; i<num; i++)    {   ave += p_0[i];    }
  ave=  ave/num;      // Calc. Mean 
  
  par[0]= ave;  par[1]= 0.2;  par[2]= 0.2;   par[3]= 3.0;
}

// Fitting exponential function via Levenberg-Marquardt method: Least Square
double  solv_LM( int num, const double x[], const double p_0[], double *c_LM, double par[], const double q[] )
{ 
  int     i, j;      long    n= DIM, inc=1, info=1; 
  long    piv[n];    double  F_x[n], J_x[n*n], err_o;  
  
  err_o= calc_FH( num, x, p_0, par, q, F_x, J_x);
  
  for (i= 0; i<n; i++)  // Modified Hessian: LM-method
    {
      J_x[ (n+1)*i ]= (1+ (*c_LM) )* J_x[ (n+1)*i ];          // Diagonal part
      for (j= i+1; j<n; j++)     J_x[ i+n*j ]= J_x[ j+n*i ];  // Non-Diagonal part
    }
  
  // Solve Likelihood Eq. 
  dgesv_( &n, &inc, J_x, &n, piv, F_x, &n, &info);           

  double  p_o[n], err_n= 0, eps= 1, t_min= 0.5, t_max= 20.0;  
  if ( info==0 )  // Parameter Optimization 
    { 
      for(i=0; i<n; i++)   
	{  p_o[i]= par[i];   par[i]+= F_x[i];  }  	
      
      // -1< r< 1
      if ( par[1]< -1 )   par[1]= -1;
      if ( par[1]>  1 )   par[1]=  1;
      // t_min < tau < t_max      
      if ( par[3]< t_min )   par[3]= t_min;
      if ( par[3]> t_max )   par[3]= t_max;
      
      // Calc. Likelihood using new parameters. 
      err_n= calc_FH( num, x, p_0, par, q, F_x, J_x);        
      eps= fabs( err_o- err_n);
      
      if ( err_n > err_o )   // No improvement...
	{  
	  *c_LM= 10* (*c_LM);   eps= 1;
	  for (i=0; i<n; i++)  {  par[i]= p_o[i];  }
	}
      else   {  *c_LM= (*c_LM)/10;    }
    }
  else     {  *c_LM= 10* (*c_LM);   eps= 1;   } 
  
  return eps;
}


// Calc. Gradient and Hessian, Rwturn Error. 
double calc_FH( int num, const double x[], const double p_0[], const double th[], const double q[], double z[], double H[])
{
  int     i= 1, j, k, n= DIM;  
  // Initilization: z, H
  for (i=0; i<n;   i++)   z[i]= 0;
  for (i=0; i<n*n; i++)   H[i]= 0;    
  
  // Calc. Gradient: F_x, Hessian: J_x. 
  double  err, f_x, t_exp, t_sin, t_cos, w[DIM], sum[2]= {0};
  for (i=0; i<num; i++)    
    { // q: constants, th: fitted parameters
      t_exp= exp( (-1)*x[i]/ th[3] );                 
      t_sin= sin( q[1]* (x[i]+q[0]+ th[2]) );   
      t_cos= cos( q[1]* (x[i]+q[0]+ th[2]) );   
      f_x=   th[0]* (1.0- th[1]* t_sin)* t_exp;      
      err= p_0[i]- f_x;      
      
      // Set w[*]: Derivative of p(t)
      w[0]= f_x/ th[0];      w[1]= (-1)* th[0]* t_sin* t_exp;
      w[2]= (-1)* q[1]* th[0]* th[1]* t_cos* t_exp;    
      w[3]= f_x* x[i]/ (th[3]* th[3]);
      
      // z_k = ¥sum err* df/d¥theta_k 
      for (j=0; j<n;j++)          {  z[j] += err* w[j];   }
      // H_{jk} = df/d¥theta_j * df/d¥theta_k 
      for (j=0; j<n; j++)  
	{
	  for (k=0; k<= j; k++)   {  H[j+k*n] += w[j]* w[k];   }
	}
      
      sum[0] += fabs(err);     // sum[0]: Error
    }
  
  return   sum[0];
}
